(* ******************************************************************** *)
(*
 * Project         : HOL-CSP - A Shallow Embedding of CSP in  Isabelle/HOL
 * Version         : 2.0
 *
 * Author          : Burkhart Wolff
 *                   (Based on HOL-CSP 1.0 by Haykal Tej and Burkhart Wolff)
 *
 * This file       : A Reference Example
 *
 * Copyright (c) 2009 Université Paris-Sud, France
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *
 *     * Neither the name of the copyright holders nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)
(* $Id: ROOT.ML 8591 2009-09-26 09:58:14Z brucker $ *)


header{* Example: Refinement Example with Buffer over infinite Alphabet*}

theory     CopyBuffer
imports    "../src/CSP"
begin 

section{* Defining the Copy-Buffer Example *}

datatype 'a channel = left 'a | right 'a | mid 'a | ack

definition SYN  :: "('a channel) set"
where     "SYN  \<equiv> (range mid) \<union> {ack}"

definition COPY :: "('a channel) process"
where     "COPY \<equiv> (\<mu> COPY. left`?`x \<rightarrow> right`!`x \<rightarrow> COPY)"

definition SEND :: "('a channel) process"
where     "SEND \<equiv> (\<mu> SEND. left`?`x \<rightarrow> mid`!`x \<rightarrow> ack \<rightarrow> SEND)"

definition REC  :: "('a channel) process"
where     "REC  \<equiv> (\<mu> REC. mid`?`x \<rightarrow> right`!`x \<rightarrow> ack \<rightarrow> REC)"


definition SYSTEM :: "('a channel) process"
where     "SYSTEM \<equiv> ((SEND \<lbrakk> SYN \<rbrakk> REC) \\ SYN)"


section{* The Standard Proof *}

subsection{* Channels and Synchronization Sets*}

text{* First part: abstract properties for these events to SYN.
       This kind of stuff could be automated easily by some
       extra-syntax for channels and SYN-sets. *}

lemma [simp]: "left x  \<notin>  SYN"
    by(auto simp: SYN_def)
  
lemma [simp]: "right x  \<notin>  SYN"
    by(auto simp: SYN_def)

lemma [simp]: "ack \<in> SYN"
    by(auto simp: SYN_def)

lemma [simp]: "mid x \<in> SYN"
    by(auto simp: SYN_def)

lemma [simp]: "inj mid"  
  by(auto simp: inj_on_def)


subsection{* Definitions by Recursors *}

text{* Second part: Derive recursive process equations, which
       are easier to handle in proofs. This part IS actually
       automated if we could reuse the fixrec-syntax below. *}

lemma COPY_rec:
  "(COPY::'a channel process) = (left`?`x \<rightarrow> right`!`x \<rightarrow> COPY)"
  by(simp add: COPY_def,rule trans, rule fix_eq, simp)
  
lemma SEND_rec: 
  "SEND = (left`?`x \<rightarrow> mid`!`x \<rightarrow> ack \<rightarrow> SEND)"
  by(simp add: SEND_def,rule trans, rule fix_eq, simp)

lemma REC_rec:
  "REC = (mid`?`x \<rightarrow> right`!`x \<rightarrow> ack \<rightarrow> REC)"
  by(simp add: REC_def,rule trans, rule fix_eq, simp)


subsection{* A Refinement Proof *}

text{* Third part: No comes the proof by fixpoint induction. 
       Not too bad in automation considering what is inferred,
       but wouldn't scale for large examples. *}


lemma impl_refines_spec : "(COPY::'a channel process) \<sqsubseteq> SYSTEM"
  apply(simp add: SYSTEM_def COPY_def)
  apply(rule fix_ind, simp_all)
  apply(subst SEND_rec, subst REC_rec)
  apply(simp add: sync_rules hide_rules)
  apply(simp add: mono_rules)
done

lemma spec_refines_impl : 
assumes fin:  "finite (SYN::('a channel)set)"
shows         "SYSTEM \<sqsubseteq> (COPY::'a channel process)"
  apply(unfold SYSTEM_def SEND_def)
  apply(rule fix_ind, simp_all add:  fin )
  apply(subst COPY_rec, subst REC_rec)
  apply(simp add: sync_rules hide_rules)
  apply(simp add: mono_rules)
done


text{* Note that this was actually proven for the
       Process ordering, not the refinement ordering. 
       But the former implies the latter.
       And due o anti-symmetrie, equality follows
       for the case of finite alphabets \ldots *}

lemma spec_equal_impl : 
assumes fin:  "finite (SYN::('a channel)set)"
shows         "SYSTEM = (COPY::'a channel process)"
  apply(rule Porder.po_class.below_antisym)
  apply(rule spec_refines_impl[OF fin])
  apply(rule impl_refines_spec)
done

section{* An Alternative Approach: Using the fixrec-Package *}

subsection{* Channels and Synchronisation Sets *}

text{* As before. *}

subsection{* Process Definitions via fixrec-Package  *}

fixrec
  COPY' :: "('a channel) process"
and
  SEND' :: "('a channel) process"
and
  REC' :: "('a channel) process"
where
   COPY'_rec[simp del]:  "COPY' = (left`?`x \<rightarrow> right`!`x \<rightarrow> COPY')"
|  SEND'_rec[simp del]:  "SEND' = (left`?`x \<rightarrow> mid`!`x \<rightarrow> ack \<rightarrow> SEND')"
|  REC'_rec[simp del] :  "REC'  = (mid`?`x \<rightarrow> right`!`x \<rightarrow> ack \<rightarrow> REC')"

find_theorems name:"COPY"

definition SYSTEM' :: "('a channel) process"
where     "SYSTEM' \<equiv> ((SEND' \<lbrakk> SYN \<rbrakk> REC') \\ SYN)"

subsection{* Another Refinement Proof on fixrec-infrastructure *}

text{* Third part: No comes the proof by fixpoint induction. 
       Not too bad in automation considering what is inferred,
       but wouldn't scale for large examples. *}

lemma impl_refines_spec' : "(COPY'::'a channel process) \<sqsubseteq> SYSTEM'"
  apply(unfold SYSTEM'_def)
  apply(rule_tac P="\<lambda> a b c. a \<sqsubseteq> ?P'"in COPY'_SEND'_REC'.induct)
  apply(simp_all add: split_def)
  apply(subst SEND'_rec) apply( subst REC'_rec) 
  apply(simp add: sync_rules hide_rules)
  apply(simp only: mono_rules)
done

lemma spec_refines_impl' : 
assumes fin:  "finite (SYN::('a channel)set)"
shows         "SYSTEM' \<sqsubseteq> (COPY'::'a channel process)"
  apply(unfold SYSTEM'_def,unfold SEND'_def)
  apply(rule fix_ind)
  apply(simp_all add:  fin split_def)
  apply(subst COPY'_rec,subst REC'_rec)
  apply(simp add: sync_rules hide_rules)
  apply(simp add: mono_rules)
done 

lemma spec_equal_impl' : 
assumes fin:  "finite (SYN::('a channel)set)"
shows         "SYSTEM' = (COPY'::'a channel process)"
apply(rule Porder.po_class.below_antisym)
apply(rule spec_refines_impl'[OF fin])
apply(rule impl_refines_spec')
done


(* Mutual recursion parses, but does not prove equalities \<dots>*)
(* DOES NOT WORK YET. Mailed Brian. 

fixrec
  COPY' :: "('a channel) process"
and
  SEND' :: "('a channel) process"
and
  REC' :: "('a channel) process"
where
   "COPY' = (left`?`x \<rightarrow> right`!`x \<rightarrow> COPY')"
|  "SEND' = (left`?`x \<rightarrow> mid`!`x \<rightarrow> ack \<rightarrow> SEND')"
|  "REC'  = (mid`?`x \<rightarrow> right`!`x \<rightarrow> ack \<rightarrow> REC')"
*)

(* Was originally in old theory presented as follows:

defs
  COPY_def   " COPY   ==(letrec COPY = left`?`x `->` right`!`x `->` COPY
                         in     COPY)"
  SYSTEM_def " SYSTEM ==(letrec
                                SEND = left`?`x`->`mid`!`x`->`ack`->`SEND;
                                REC  = mid`?`x`->`right`!`x`->` ack`->`REC
                         in     ((SEND `[|` SYN `|]` REC) `\\` SYN))"
*)


end
