(* ******************************************************************** *)
(*
 * Project         : HOL-CSP - A Shallow Embedding of CSP in  Isabelle/HOL
 * Version         : 2.0
 *
 * Author          : Burkhart Wolff
 *                   (Based on HOL-CSP 1.0 by Haykal Tej, Clemens Fischer 
                      and Burkhart Wolff)
 *
 * This file       : Plain Old Telephone System in HOL-CSP
 *
 * Copyright (c) 2010 Université Paris-Sud, France
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *
 *     * Neither the name of the copyright holders nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)
(* $Id: ROOT.ML 8591 2009-09-26 09:58:14Z brucker $ *)

header{* Example: Plain Old Telephone System*}

theory     POTS
imports    "../src/CSP"
begin 

section{* The Alphabet and basic types of POTS *}

datatype MtcO = Osetup | Odiscon_o
datatype MctO = Obusy  | Oalert    | Oconnect | Odiscon_t
datatype MtcT = Tbusy  | Talert    | Tconnect | Tdiscon_t
datatype MctT = Tsetup | Tdiscon_o

types
  Phones = "int"
  tcOT   = "Phones \<times> MtcO"
  ctOT   = "Phones \<times> MctO"
  tcOdT  = "Phones \<times> Phones"
  tcTT   = "Phones \<times> MtcT \<times> Phones"
  ctTT   = "Phones \<times> MctT \<times> Phones"
  dig    = "Phones \<times> Phones"

datatype channels = 
    tcO tcOT        | ctO ctOT            | tcT tcTT | ctT ctTT | 
    tcOdial tcOdT   | StartReject Phones  | EndReject Phones    |
    terminal Phones | off_hook Phones     | on_hook Phones      |
    digits dig      | tone_ring Phones    | tone_quiet Phones   |
    tone_busy Phones| tone_dial Phones    | connected Phones

typ tcOT

definition min_phones :: int
  where   "min_phones \<equiv> 1"

definition  max_phones :: int  
  where    "max_phones \<equiv> 5"

definition phones :: "Phones list"
  where   "phones \<equiv> [min_phones .. max_phones]"

definition EventsIPhone :: "Phones \<Rightarrow> channels set"
  where   "EventsIPhone u \<equiv> {tone_ring u, tone_quiet u, 
                              tone_busy u, tone_dial u, connected u}"
definition EventsUser :: "Phones \<Rightarrow> channels set"
  where   "EventsUser u \<equiv> {off_hook u, on_hook u} \<union>
                            {x . \<exists> n. x = digits (u,n)}"


section {* CSP_M *}

definition
  alfa_par:: "['a process,'a set, 'a set, 'a process] \<Rightarrow> 'a process"
                                        ("(3_\<lbrakk>_||_\<rbrakk>_)" 14)
where  "P \<lbrakk>A||B\<rbrakk> Q \<equiv> (P\\(- A)) \<lbrakk> A \<inter> B \<rbrakk> (Q\\(- B))"

consts
  link_par:: "['a process,'b \<Rightarrow> 'a, 'b \<Rightarrow> 'a, 'a process] \<Rightarrow> 'a process"
                                        ("(3_`[`_`<->`_`]`_)" 14)
  (* NOT YET DEFINED !!! *)
(*
  rep_link_par :: "['b \<Rightarrow> 'a, 'b \<Rightarrow> 'a, 'b list, 'b \<Rightarrow>'a process]\<Rightarrow>'a process"
  "rep_link_par c1 c2 S P \<equiv>
   foldl (%r a. r `[`c1 `<->` c2`]` (P a)) (hd S) (tl S)"
*)
syntax
  "_rep_link_par" :: "['c,'c,pttrn,'a list, 'b process] \<Rightarrow> 'b process"
                  ("(3\<lbrakk>_`<->`_\<rbrakk>_\<in>_. / _)" 14)


definition
  guard   :: "[bool, 'a process] \<Rightarrow> 'a process"        ("(2_`&`_)" [26,27]26)
where
  "a `&` P \<equiv> if a then P else STOP"

definition
  timeout    :: "['a process,'a process] \<Rightarrow> 'a process"("(2_`[>`_)"[23,22]22)
where
  "P `[>` Q \<equiv> (P \<box> Q) \<sqinter> Q"

find_theorems name:cont 
lemma cont_timeout [simp]:
  "\<lbrakk>cont f; cont g\<rbrakk> \<Longrightarrow> cont (\<lambda>x. f x `[>` g x)"
  by(simp add:timeout_def Ndet.ndet_cont Det.det_cont)

subsection {* CSP_M: Repetitive Combinators *}


subsubsection {* External Choice Repetition *}

definition
  rep_ech :: "['a set, 'a \<Rightarrow> 'b process] \<Rightarrow> 'b process"
where
  "rep_ech A P \<equiv> Finite_Set.fold (\<lambda>a r. r \<box> P a) (STOP) (A)"
syntax
  "_rep_ech"   :: "[pttrn,'a set,'b process] \<Rightarrow> 'b process"
                  ("(3\<box>_\<in>_. / _)" 15)
translations  "\<box> p \<in> A. P "       == "CONST rep_ech A (\<lambda>p. P)"


lemma rep_ech_mt[simp]: "(\<box> p \<in> {}. P p) = STOP"
by(simp add: rep_ech_def)

lemma rep_ech_insert[simp]:
assumes fin: "finite X"
shows  "(\<box> p \<in> (insert a X). P p) = (P a \<box> (\<box> p \<in> X - {a}. P p))"
apply(simp add: rep_ech_def)
apply(subst Finite_Set.comp_fun_commute.fold_insert_remove)
apply(auto simp: comp_fun_commute_def det_commute det_assoc fin comp_def )
apply(rule ext, simp add:det_assoc[symmetric]det_left_commute )
apply(simp only: det_assoc)
apply(subst det_commute) back apply simp
done


lemma test: "(\<box> p\<in>{1::int .. 3}. P p) = (P 1 \<box> P 2 \<box> P 3)"
by(simp add: det_assoc Presburger.simp_from_to)

lemma rep_ech_cont[simp]:
assumes fin: "finite A"
assumes con: "(\<And>x. x\<in>A \<Longrightarrow> cont  (P x))"
shows   "cont (\<lambda>y. \<box> z\<in>A. P z y)"
apply(rule Finite_Set.finite_subset_induct[OF fin], auto simp:con)
sorry

(* worked under 2011 with:

lemma rep_ech_cont[simp]:
assumes fin: "finite A"
assumes con: "(\<And>x. x\<in>A \<Longrightarrow> cont  (P x))"
shows   "cont (\<lambda>y. \<box> z\<in>A. P z y)"
by(rule Finite_Set.finite_subset_induct[OF fin], auto simp:con)

*)

lemma rep_ech_bot[rule_format]:
assumes fin: "finite A"
assumes dom: "a \<in> A"
assumes bot: "P a = \<bottom>"
shows   "(\<box> x\<in>A. P x) = \<bottom>"
apply(rule rev_mp[OF dom], rule rev_mp[OF bot])
by(rule Finite_Set.finite_subset_induct[OF fin], auto)


subsubsection {* Internal Choice Repetition *}

definition
  rep_ich :: "['a set, 'a \<Rightarrow> 'b process] \<Rightarrow> 'b process"
where
 "rep_ich A P \<equiv> Finite_Set.fold1 (op \<sqinter>) (P ` A)"

syntax
 "_rep_ich"   :: "[pttrn,'a set,'b process] \<Rightarrow> 'b process"
                  ("(3\<sqinter> _\<in>_. / _)" 17)
translations  "\<sqinter> p \<in> A. P "   == "CONST rep_ich A (\<lambda>p. P)"


subsubsection {* Sequential Repetition *}

definition
  rep_seq :: "['a list, 'a \<Rightarrow> 'b process] \<Rightarrow> 'b process"
where
  "rep_seq S P \<equiv> foldl (%r a. r `;` P a) (SKIP) (S)"
syntax
  "_rep_seq"   :: "[pttrn,'a set,'a process] \<Rightarrow> 'a process"
                  ("(3`;`_\<in>_ \<rightarrow>/ _)" 24)
translations  "`;` p \<in> A \<rightarrow> P "    == "CONST rep_seq A (\<lambda>p. P)"




subsubsection {* Interleave Repetition *}

definition
  rep_int :: "['a list, 'a \<Rightarrow> 'b process] \<Rightarrow> 'b process"
where
 "rep_int S P \<equiv> foldl (\<lambda>r a. r ||| P a) (SKIP) (S)"
syntax
 "_rep_int"   :: "[pttrn,'a list, 'b process] \<Rightarrow> 'b process"
                  ("(3|||_\<in>_. / _)" 11)
translations  "||| p \<in> A. P "  == "CONST rep_int A (\<lambda>p. P)"


subsubsection {* Synchronized Parallel Repetition *}

definition
  rep_shint :: "['b set, 'a list, 'a \<Rightarrow> 'b process] \<Rightarrow> 'b process"
where
  "rep_shint A S P \<equiv> foldl (\<lambda>r a. r \<lbrakk>A\<rbrakk> (P a)) (SKIP) (S)"
syntax
  "_rep_shint" :: "['b set,pttrn,'a list, 'b process] \<Rightarrow> 'b process"
                  ("(3\<lbrakk>_\<rbrakk>_\<in>_. / _)" 14)

(*weird stuff *)
translations  "\<lbrakk>c1`<->`c2\<rbrakk> p \<in> A. P" == "CONST rep_shint c1 c2 A (\<lambda>p. P)"
translations  "\<lbrakk>B\<rbrakk> p \<in> A. P"       == "CONST rep_shint B A (\<lambda>p. P)"

definition
  CHAOS  :: "'a set \<Rightarrow> 'a process"
where
  "CHAOS A \<equiv> (\<mu> X . (rep_ech A (\<lambda>x. x \<rightarrow> X)) \<sqinter> STOP)"



definition   refines_fd :: "['a process,'a process] \<Rightarrow> bool" ("_ `[FD=` _" [53,52]52)
where   "P `[FD=` Q \<equiv>  (P << Q)"
definition   refines_f :: "['a process,'a process] \<Rightarrow> bool" ("_ `[F=` _" [53,52]52)
where   "P `[F=` Q \<equiv>  (F Q <= F P)"
definition   refines_t :: "['a process,'a process] \<Rightarrow> bool" ("_ `[T=` _" [53,52]52)
where   "P `[T=` Q \<equiv>  (T Q <= T P)"


(* TO BE PORTED !
local
fun print_readX [chan,Const ("Collect",_) $ set,proc] =
    let fun strip_split ((Const ("split", _)$ Abs(c,tt,t))) =
            let val (s,R) = strip_split t
            in  ((c,tt)::s,R) end
          | strip_split (Abs(c,tt,t)) = ([(c,tt)],t);
        fun mk_tuple (a :: b :: R) = Const ("Pair",dummyT) $
                                   a $ (mk_tuple (b :: R))
           |mk_tuple (a::[])       = a;
        val (s,set_base) = strip_split set;
        val (s',proc_base) = strip_split proc
        val ss = map Free s
    in  if length s = length s' then
           (Const ("_readX",dummyT)) $ chan $
           (mk_tuple ss) $
           (subst_bounds(rev ss,set_base)) $
           (subst_bounds(rev ss,proc_base))
        else raise Match
    end
   | print_readX _ = raise Match;
in
val print_translation = [("read", print_readX)];
*)

section{*Auxilliaries to Substructure the Specification*}

definition
  Tside_connected     :: "[Phones, Phones] \<Rightarrow> channels process"
where "Tside_connected ts os \<equiv> 
           (ctT`!`(ts,Tdiscon_o,os) 
            \<rightarrow> tcT`!`(ts,Tdiscon_t,os)
            \<rightarrow> EndReject`!`ts\<rightarrow>SKIP)
      `[>` (tcT`!`(ts,Tdiscon_t,os) 
            \<rightarrow>ctT`!`(ts,Tdiscon_o,os)
            \<rightarrow>EndReject`!`ts\<rightarrow>SKIP)"

definition
  Oside_connected     :: "Phones \<Rightarrow> channels process"
where   "Oside_connected ts \<equiv>
            (ctO`!`(ts,Odiscon_t)
             \<rightarrow>tcO`!`(ts,Odiscon_o)
             \<rightarrow>EndReject`!`ts\<rightarrow>SKIP)
        `[>`(tcO`!`(ts,Odiscon_o)
             \<rightarrow>ctO`!`(ts,Odiscon_t)
             \<rightarrow>EndReject`!`ts\<rightarrow>SKIP)"

definition
  Oside1              :: "[Phones, Phones] \<Rightarrow> channels process"
where 
  "Oside1 ts p \<equiv> 
             tcOdial`!` (ts,p)
	     \<rightarrow>(  (ctO`!`(ts,Oalert)
                     \<rightarrow>ctO`!`(ts,Oconnect)
                     \<rightarrow>(Oside_connected ts))
                 \<box>(ctO`!`(ts,Oconnect)\<rightarrow>(Oside_connected ts))
                 \<box>(ctO`!`(ts,Obusy)\<rightarrow>tcO`!`(ts,Odiscon_o)
                    \<rightarrow>EndReject`!`ts\<rightarrow>SKIP))"

definition
  ITside_connected    :: "[Phones,Phones,channels process] \<Rightarrow> channels process"
where
 "ITside_connected ts os IT \<equiv>

           ((ctT(ts,Tdiscon_o,os)
           \<rightarrow>(   (tone_busy`!`ts\<rightarrow>on_hook`!`ts
                   \<rightarrow>tcT`!`(ts,Tdiscon_t,os)\<rightarrow>EndReject`!`ts\<rightarrow> IT)
               \<box>(on_hook`!`ts\<rightarrow>tcT`!`(ts,Tdiscon_t,os)
                    \<rightarrow>EndReject`!`ts\<rightarrow>IT)
               ))
            \<box>(on_hook`!`ts\<rightarrow>tcT`!`(ts,Tdiscon_t,os)
                 \<rightarrow>ctT`!`(ts,Tdiscon_o,os)\<rightarrow>EndReject`!`ts\<rightarrow>IT)
            )"


   
section{*A Telephone *}

(* THIS DOES NOT WORK - NO MIXTURE BETWEEN 
   standard and continuous function space possible
   in the fixrec- package.
fixrec
  T :: "Phones \<Rightarrow> channels process"
and
  Oside :: "Phones \<Rightarrow> channels process"
and
  Tside :: "channels process"
where
   T_rec[simp del]:    "T = (\<lambda> ts. ((Tside `;` (T ts))
                                  `[>` (Oside ts `;` (T ts))))"
|  Oside_rec[simp del]:"Oside = (\<lambda> ts. (StartReject`!`ts \<rightarrow>
                                 tcO`!`(ts,Osetup) \<rightarrow>
    	                          (SKIP)))"
|  Tside_rec[simp del]:  "Tside  = SKIP"
*)
(*
fixrec
  T :: "Phones \<Rightarrow> channels process"
and
  Oside :: "Phones \<Rightarrow> channels process"
and
  Tside :: "channels process"
where
   T_rec[simp del]:    "T = (\<lambda> ts. ((Tside `;` (T ts))
                                  `[>` (Oside ts `;` (T ts))))"
|  Oside_rec[simp del]:"Oside = (\<lambda> ts. (StartReject`!`ts \<rightarrow>
                                 tcO`!`(ts,Osetup) \<rightarrow>
    	                          (\<sqinter> p \<in>(set phones). Oside1 ts p)))"
|  Tside_rec[simp del]:  "Tside  = SKIP"
*)

text{* The standard fixrec-package does not work here since there are
  too complex mixtures between standard function space and continouos
  function space necessary. For this reason, we make the development
  by foot. *}


definition
  T_Oside_Tside_NoReject_Reject_REC ::
       "Phones \<Rightarrow> (channels process \<times> channels process \<times> channels process \<times>
                   channels process \<times> channels process)"
where "T_Oside_Tside_NoReject_Reject_REC ts \<equiv> undefined"
  

(* bu, 11.10.2010: COULD NOT PORT THE FOLLOWING TO Isabelle2009-2 
definition
  T_Oside_Tside_NoReject_Reject_REC ::
       "Phones \<Rightarrow> (channels process \<times> channels process \<times> channels process \<times>
                   channels process \<times> channels process)"
where "T_Oside_Tside_NoReject_Reject_REC ts \<equiv>
       (Letrec T = Skip ;
              Oside = Skip ;
              Tside = Skip ;
              NoReject = Skip ;
              Reject = Skip 
       in (T , T , T , T , T))"
  

definition
  T_Oside_Tside_NoReject_Reject_REC ::
       "Phones \<Rightarrow> (channels process \<times> channels process \<times> channels process \<times>
                   channels process \<times> channels process)"
where 
  "T_Oside_Tside_NoReject_Reject_REC ts \<equiv> fix \<cdot>
      (  \<Lambda> \<langle>T, Oside, Tside, NoReject, Reject\<rangle>.
            (  ((Tside `;` T) `[>` (Oside`;` T )),
               (StartReject `!` ts \<rightarrow> tcO `!` (ts,Osetup) \<rightarrow>
    	                (\<sqinter> p \<in>(set phones). Oside1 ts p)),
               (ctT `?` (y,z,os) `|` ((y,z)=(ts,Tsetup)) \<rightarrow>
                       StartReject `!` ts \<rightarrow>
                      (   (tcT `!` (ts,Talert,os)
                           \<rightarrow>tcT `!` (ts,Tconnect,os)
                           \<rightarrow>(Tside_connected ts os))
                       \<sqinter> tcT `!` (ts,Tconnect,os)
                             \<rightarrow> (Tside_connected ts os))),
               (StartReject `!` ts \<rightarrow> Reject),
               ((ctT `?` (y,z,os)`|` ((y,z)=(ts,Tsetup) \<and>
                              os \<in> {p. p mem phones \<and> p\<noteq>ts})
			    \<rightarrow> tcT `!` (ts,Tbusy,os) \<rightarrow> Reject) \<box>
               (EndReject `!` ts\<rightarrow>NoReject))
             )
      )"

*)

definition T::"Phones \<Rightarrow> channels process"
where "T ts \<equiv> fst (T_Oside_Tside_NoReject_Reject_REC ts)"

definition Oside::"Phones \<Rightarrow> channels process"
where "Oside ts \<equiv> fst (snd(T_Oside_Tside_NoReject_Reject_REC ts))"

definition Tside::"Phones \<Rightarrow> channels process"
where "Tside ts \<equiv> fst (snd(snd(T_Oside_Tside_NoReject_Reject_REC ts)))"

definition NoReject::"Phones \<Rightarrow> channels process"
where "NoReject ts \<equiv> fst (snd(snd(snd(T_Oside_Tside_NoReject_Reject_REC ts))))"

definition Reject::"Phones \<Rightarrow> channels process"
where "Reject ts \<equiv> fst (snd(snd(T_Oside_Tside_NoReject_Reject_REC ts)))"

definition Tel:: "Phones \<Rightarrow> channels process"
where "Tel p \<equiv> (T p \<lbrakk>{StartReject p, EndReject p}\<rbrakk> NoReject p)
	        \\  {StartReject p, EndReject p}"

lemma hack: "!! f. cont f" sorry (* we don't consider continuity here \<dots> *)
lemma T_rec :
     "T ts = ((Tside ts `;` T ts) `[>` (Oside ts`;` T ts))"
apply(subst T_def, subst T_Oside_Tside_NoReject_Reject_REC_def)

(* bu, 11.10.2010: COULD NOT PORT THE FOLLOWING TO Isabelle2009-2 

apply(subst fix_eq)
apply(fold T_Oside_Tside_NoReject_Reject_REC_def)
apply(simp add: Cprod.csplit_def)
apply(rule_tac P="%x. fst(x) = ?Y" in beta_cfun[THEN ssubst, OF hack])
apply(simp_all add: Cprod.csplit_def beta_cfun[OF hack]
                   Cprod.cfst_def Cprod.csnd_def)
apply(fold Tside_def Oside_def POTS.T_def, simp)
done

lemma Oside_rec :
     "Oside ts = (StartReject`!`ts \<rightarrow> tcO`!`(ts,Osetup) \<rightarrow>
    	                (\<sqinter> p \<in>(set phones). Oside1 ts p))"
sorry

lemma Tside_rec :
     "Tside ts = (ctT`?`(y,z,os)`|`((y,z)=(ts,Tsetup)) \<rightarrow>
                       StartReject`!`ts \<rightarrow>
                      (   (tcT`!`(ts,Talert,os)
                           \<rightarrow>tcT`!`(ts,Tconnect,os)
                           \<rightarrow>(Tside_connected ts os))
                       \<sqinter> tcT`!`(ts,Tconnect,os)
                             \<rightarrow> (Tside_connected ts os)))"
sorry

lemma NoReject_rec:
     "NoReject ts =  (StartReject`!`ts \<rightarrow> Reject ts)"
sorry

lemma Reject_rec :
     "Reject ts = ((ctT`?`(y,z,os)`|`((y,z)=(ts,Tsetup) \<and>
                              os \<in> {p. p mem phones \<and> p\<noteq>ts})
			    \<rightarrow> tcT`!`(ts,Tbusy,os) \<rightarrow> Reject ts) \<box>
                          (EndReject`!`ts\<rightarrow> NoReject ts))"
sorry


(* This was originally in the HOL-CSP 1.0 Version:
defs
 Tel_def   
   "Tel ts \<equiv> letrec
 
    	         T = ((Tside `;` T) `[>` (Oside `;` T));

	
   	         Oside = StartReject`!`ts \<rightarrow> tcO`!`(ts,Osetup)\<rightarrow>
    	                (\<sqinter>p `:` set phones`@`Oside1 ts p);


      	         Tside = ctT`?`(y,z,os)`|`((y,z)=(ts,Tsetup))
                          \<rightarrow>StartReject`!`ts\<rightarrow>
                      (   (tcT`!`(ts,Talert,os)
                           \<rightarrow>tcT`!`(ts,Tconnect,os)
                           \<rightarrow>(Tside_connected ts os))
                       \<sqinter> tcT`!`(ts,Tconnect,os)
                             \<rightarrow>(Tside_connected ts os));


                 NoReject = StartReject`!`ts\<rightarrow>Reject;


                 Reject   = (ctT`?`(y,z,os)`|`((y,z)=(ts,Tsetup)&
				  	  os:{p. p mem phones & p~=ts}) 
			    \<rightarrow> tcT`!`(ts,Tbusy,os) \<rightarrow> Reject)
                            \<box>(EndReject`!`ts\<rightarrow>NoReject)

             in  (T \<lbrakk>{StartReject ts, EndReject ts}\<rbrakk> NoReject)
	         `\\`{StartReject ts, EndReject ts}"

*)


section{* A Connector with the Network *}

definition
  Call_BUSY_Connected_REC ::
       "Phones \<Rightarrow> (channels process \<times>
                   (Phones \<Rightarrow> channels process) \<times>
                   (Phones \<Rightarrow> channels process))"
where
  "Call_BUSY_Connected_REC os \<equiv>
   (fix \<cdot>
      (\<Lambda>\<langle>Call, BUSY, Connected\<rangle>.
           ( ((tcO `!` (os,Osetup)
                     \<rightarrow> tcOdial`?`(x,ts)`|`(x=os)
                     \<rightarrow> (BUSY ts))`;`Call),
             (\<lambda>ts. if ts = os then ctO`!`(os,Obusy)
                                       \<rightarrow> tcO`!`(os,Odiscon_o) \<rightarrow> SKIP
                          else  ctT`!`(ts,Tsetup,os)
                                \<rightarrow>( (tcT`!`(ts,Tbusy,os)
                                       \<rightarrow> ctO`!`(os,Obusy)
                                       \<rightarrow> tcO`!`(os,Odiscon_o) \<rightarrow> SKIP)
                                    \<box>
                                      (tcT`!`(ts,Talert,os)
                                       \<rightarrow> ctO`!`(os,Oalert)
                                       \<rightarrow> tcT`!`(ts,Tconnect,os)
                                       \<rightarrow> ctO`!`(os,Oconnect)
                                       \<rightarrow> Connected ts)
                                    \<box>
                                      (tcT`!`(ts,Tconnect,os)
                                       \<rightarrow> ctO`!`(os,Oconnect)
                                       \<rightarrow> Connected ts))),
             (\<lambda>ts. (tcO`!`(os,Odiscon_o) \<rightarrow>
                             (( (ctT`!`(ts,Tdiscon_o,os)
                                 \<rightarrow> tcT`!`(ts,Tdiscon_t,os)
                                 \<rightarrow> SKIP)
                              \<box>
                                (tcT`!`(ts,Tdiscon_t,os)
                                 \<rightarrow> ctT`!`(ts,Tdiscon_o,os)
                                 \<rightarrow> SKIP)
                              ) `;` (ctO`!`(os,Odiscon_t) \<rightarrow> SKIP)))
                          \<box>
                             (tcT`!`(ts,Tdiscon_t,os) \<rightarrow>
                             (  (  ctO`!`(os,Odiscon_t)
                                 \<rightarrow> ctT`!`(ts,Tdiscon_o,os)
                                 \<rightarrow> tcO`!`(os,Odiscon_o) \<rightarrow> SKIP )
                               \<box>
                                (tcO`!`(os,Odiscon_o)
                                 \<rightarrow> ctT`!`(ts,Tdiscon_o,os)
                                 \<rightarrow> ctO(os,Odiscon_t)\<rightarrow> SKIP) )
                             )
                          )
            )
      ))"

definition Call :: "Phones \<Rightarrow> channels process"
where     "Call os \<equiv> fst(Call_BUSY_Connected_REC os)"

definition BUSY :: "Phones \<Rightarrow> Phones \<Rightarrow> channels process"
where     "BUSY os ts \<equiv> fst(snd(Call_BUSY_Connected_REC os)) ts"

definition Connected :: "Phones \<Rightarrow> Phones \<Rightarrow> channels process"
where     "Connected os ts \<equiv> snd(snd(Call_BUSY_Connected_REC os)) ts"


lemma Call_rec :
     "Call os = ((tcO`!`(os,Osetup)
                          \<rightarrow> tcOdial`?`(x,ts)`|`(x=os)
                          \<rightarrow> (BUSY os ts))`;`Call os)"
sorry

lemma BUSY_rec :
     "BUSY os ts = (if ts = os then ctO`!`(os,Obusy)
                                       \<rightarrow> tcO`!`(os,Odiscon_o) \<rightarrow> SKIP
                          else  ctT`!`(ts,Tsetup,os)
                                \<rightarrow>( (tcT`!`(ts,Tbusy,os)
                                       \<rightarrow> ctO`!`(os,Obusy)
                                       \<rightarrow> tcO`!`(os,Odiscon_o) \<rightarrow> SKIP)
                                    \<box>
                                      (tcT`!`(ts,Talert,os)
                                       \<rightarrow> ctO`!`(os,Oalert)
                                       \<rightarrow> tcT`!`(ts,Tconnect,os)
                                       \<rightarrow> ctO`!`(os,Oconnect)
                                       \<rightarrow> Connected os ts)
                                    \<box>
                                      (tcT`!`(ts,Tconnect,os)
                                       \<rightarrow> ctO`!`(os,Oconnect)
                                       \<rightarrow> Connected os ts)))"
sorry

lemma "Connected os ts = ((tcO`!`(os,Odiscon_o) \<rightarrow>
                             (( (ctT`!`(ts,Tdiscon_o,os)
                                 \<rightarrow> tcT`!`(ts,Tdiscon_t,os)
                                 \<rightarrow> SKIP)
                              \<box>
                                (tcT`!`(ts,Tdiscon_t,os)
                                 \<rightarrow> ctT`!`(ts,Tdiscon_o,os)
                                 \<rightarrow> SKIP)
                              ) `;` (ctO`!`(os,Odiscon_t) \<rightarrow> SKIP)))
                          \<box>
                             (tcT`!`(ts,Tdiscon_t,os) \<rightarrow>
                             (  (  ctO`!`(os,Odiscon_t)
                                 \<rightarrow> ctT`!`(ts,Tdiscon_o,os)
                                 \<rightarrow> tcO`!`(os,Odiscon_o) \<rightarrow> SKIP )
                               \<box>
                                (tcO`!`(os,Odiscon_o)
                                 \<rightarrow> ctT`!`(ts,Tdiscon_o,os)
                                 \<rightarrow> ctO(os,Odiscon_t)\<rightarrow> SKIP) )
                             )
                          )"
sorry

(* This has been in HOL-CSP 1.0 :
defs

Call_def
  "Call(os)  \<equiv> 
    (letrec C    = ((tcO`!`(os,Osetup) 
                     \<rightarrow> tcOdial`?`(x,ts)`|`(x=os)
                     \<rightarrow> (BUSY ts))`;`C);
                                                                            
            BUSY  = (%ts. if ts = os then ctO`!`(os,Obusy) 
                                       \<rightarrow> tcO`!`(os,Odiscon_o) \<rightarrow> SKIP
                          else  ctT`!`(ts,Tsetup,os)                          
                                \<rightarrow>( (tcT`!`(ts,Tbusy,os)     
                                       \<rightarrow> ctO`!`(os,Obusy)
                                       \<rightarrow> tcO`!`(os,Odiscon_o) \<rightarrow> SKIP)
                                    \<box>
                                      (tcT`!`(ts,Talert,os)        
                                       \<rightarrow> ctO`!`(os,Oalert)
                                       \<rightarrow> tcT`!`(ts,Tconnect,os)
                                       \<rightarrow> ctO`!`(os,Oconnect)
                                       \<rightarrow> Connected ts)
                                    \<box>
                                      (tcT`!`(ts,Tconnect,os)   
                                       \<rightarrow> ctO`!`(os,Oconnect)
                                       \<rightarrow> Connected ts)));
                                                                                       
           Connected  = (%ts.(tcO`!`(os,Odiscon_o) \<rightarrow>
                             (( (ctT`!`(ts,Tdiscon_o,os) 
                                 \<rightarrow> tcT`!`(ts,Tdiscon_t,os)
                                 \<rightarrow> SKIP)
                              \<box>
                                (tcT`!`(ts,Tdiscon_t,os) 
                                 \<rightarrow> ctT`!`(ts,Tdiscon_o,os)
                                 \<rightarrow> SKIP)
                              ) `;` (ctO`!`(os,Odiscon_t) \<rightarrow> SKIP)))
                          \<box>
                             (tcT`!`(ts,Tdiscon_t,os) \<rightarrow>
                             (  (  ctO`!`(os,Odiscon_t) 
                                 \<rightarrow> ctT`!`(ts,Tdiscon_o,os)
                                 \<rightarrow> tcO`!`(os,Odiscon_o) \<rightarrow> SKIP )
                               \<box>
                                (tcO`!`(os,Odiscon_o) 
                                 \<rightarrow> ctT`!`(ts,Tdiscon_o,os)
                                 \<rightarrow> ctO(os,Odiscon_t)\<rightarrow> SKIP) )
                             ) 
                          )                                                         
    in C)"

*)

section{* Combining NETWORK and TELEPHONES to a SYSTEM *}

text{* Parameter: *}

consts      VisibleEvents :: "channels set"

definition NETWORK     :: "channels process"
where     "NETWORK     \<equiv> ( ||| os \<in> phones. Call os)"

definition  TELEPHONES :: "channels process"
where      "TELEPHONES \<equiv> ( ||| ts \<in> phones. Tel ts)"

definition  SYSTEM     :: "channels process"
where      "SYSTEM     \<equiv> NETWORK \<lbrakk>VisibleEvents\<rbrakk> TELEPHONES"


section{*A simple model of a user *}

definition
  User_UserSCon_REC :: "Phones \<Rightarrow> (channels process \<times> channels process)"
where 
  "User_UserSCon_REC u \<equiv> fix \<cdot>
      (\<Lambda> \<langle>User, UserSCon\<rangle>.
            (((off_hook`!`u \<rightarrow>
             (  (tone_dial`!`u \<rightarrow>
                     (\<sqinter> p \<in> set phones. digits`!`(u,p)\<rightarrow>tone_quiet`!`u\<rightarrow>
                                     (  (tone_ring`!`u\<rightarrow>connected`!`u\<rightarrow>UserSCon)
                                      \<box>(connected`!`u\<rightarrow>UserSCon)
                                      \<box>(tone_busy`!`u\<rightarrow>on_hook`!`u\<rightarrow>User)
                                     )
                     )
                 )
                 \<box> connected`!`u \<rightarrow> UserSCon
              ))
             \<box> tone_ring`!`u\<rightarrow>off_hook`!`u\<rightarrow>connected`!`u \<rightarrow>UserSCon),
            ((tone_busy`!`u\<rightarrow>on_hook`!`u\<rightarrow>User)
                 `[>` on_hook`!`u\<rightarrow>User)
           )
       )"



definition User :: "Phones \<Rightarrow> channels process"
where "User u \<equiv> fst(User_UserSCon_REC u)"

definition UserSCon :: "Phones \<Rightarrow> channels process"
where "UserSCon u \<equiv> snd(User_UserSCon_REC u)"

(* was in HOL-CSP 1.0:
"UserS u \<equiv>
  letrec
     User = ((off_hook`!`u \<rightarrow>
             (  (tone_dial`!`u \<rightarrow>
                     (\<sqinter>p `:` set phones`@`digits`!`(u,p)\<rightarrow>tone_quiet`!`u\<rightarrow>
                                     (  (tone_ring`!`u\<rightarrow>connected`!`u\<rightarrow>UserSCon)
                                      \<box>(connected`!`u\<rightarrow>UserSCon)
                                      \<box>(tone_busy`!`u\<rightarrow>on_hook`!`u\<rightarrow>User)
                                     )
                     )
                 )
                 \<box> connected`!`u \<rightarrow> UserSCon
              ))
             \<box> tone_ring`!`u\<rightarrow>off_hook`!`u\<rightarrow>connected`!`u \<rightarrow>UserSCon
          );
     UserSCon =  (tone_busy`!`u\<rightarrow>on_hook`!`u\<rightarrow>User)
                 `[>` on_hook`!`u\<rightarrow>User
  in User"
*)


section{* A Telephone Implementation *}

consts
  ITel                :: "Phones \<Rightarrow> channels process"

(* NOT YET PORTED
   
"ITel ts \<equiv> 

   letrec 

       Tel =  (IT \<lbrakk>{StartReject ts,EndReject ts}\<rbrakk> NoReject)
              `\\` {StartReject ts,EndReject ts};


       NoReject = StartReject`!`ts\<rightarrow>Reject;


       Reject   = (ctT`?`(y,z,os)`|`((y,z)=(ts,Tsetup)&os:{p. p mem phones&p~=ts}) 
		  \<rightarrow> tcT`!`(ts,Tbusy,os) \<rightarrow> Reject)
             \<box>(EndReject`!`ts\<rightarrow>NoReject);


       IT = (IOside \<box> ITside);


       IOside = off_hook`!`ts
                \<rightarrow>StartReject`!`ts
                \<rightarrow>tcO`!`(ts,Osetup)
                \<rightarrow>tone_dial`!`ts
                \<rightarrow> ( digits`?`(x,os)`|`(x=ts)
                       \<rightarrow>tone_quiet`!`ts
                       \<rightarrow>tcOdial`!`(ts,os)
                       \<rightarrow> ( (ctO`!`(ts,Oalert)
                               \<rightarrow>ctO`!`(ts,Oconnect)
                               \<rightarrow>connected`!`ts
                               \<rightarrow>IOside_connected)
                            \<box> (ctO`!`(ts,Oconnect)
                                  \<rightarrow>connected`!`ts
                                  \<rightarrow>IOside_connected)
                            \<box> (ctO`!`(ts,Obusy)
                                  \<rightarrow>tone_busy`!`ts
                                  \<rightarrow>on_hook`!`ts
                                  \<rightarrow> tcO`!`(ts,Odiscon_o)
                                  \<rightarrow> EndReject`!`ts
                                  \<rightarrow>IT)
                             )
                      );

  
       IOside_connected = (ctO`!`(ts,Odiscon_t)
                           \<rightarrow> (  (tone_busy`!`ts
                                    \<rightarrow>on_hook`!`ts
                                    \<rightarrow>tcO`!`(ts,Odiscon_o)
                                    \<rightarrow>EndReject`!`ts\<rightarrow>IT)
                                \<box>(on_hook`!`ts
                                     \<rightarrow>tcO`!`(ts,Odiscon_o)
                                     \<rightarrow>EndReject`!`ts\<rightarrow>IT)
                                )
                           )

                          \<box>
                          (on_hook`!`ts
                           \<rightarrow>tcO`!`(ts,Odiscon_o)
                           \<rightarrow>ctO`!`(ts,Odiscon_t)
                           \<rightarrow> EndReject`!`ts
                           \<rightarrow>IT);

       ITside = ctT`?`(x,y,os)`|`((x,y)=(ts,Tsetup))
                \<rightarrow>StartReject`!`ts
                \<rightarrow>(   (tone_ring`!`ts
                         \<rightarrow>tcT`!`(ts,Talert,os)
                         \<rightarrow>off_hook`!`ts
                         \<rightarrow>connected`!`ts
                         \<rightarrow>tcT`!`(ts,Tconnect,os)
                         \<rightarrow> ITside_connected ts os IT)
                    \<box>(off_hook`!`ts
                         \<rightarrow> connected`!`ts
                         \<rightarrow> tcT`!`(ts,Tconnect,os)
                         \<rightarrow> ITside_connected ts os IT)
                     )


 in  Tel"
*)



definition
  ImplementT          :: "Phones \<Rightarrow> channels process"
where
"ImplementT ts \<equiv> ((ITel ts) \<lbrakk>EventsIPhone ts \<union> EventsUser ts\<rbrakk> (User ts))
                 \\ (EventsIPhone ts \<union> EventsUser ts)"


section {* Toplevel Proof-Goals*}

text{* This has been proven in ancient FDR model \<dots> *}
(*
assert !p. Tel(p) [deadlock free]
assert TELEPHONES deadlock free
assert Call(p) deadlock free
assert NETWORK deadlock free
assert SYSTEM livelock free
assert SYSTEM deadlock free

assert ImplementT : [livelock free]
assert ITel : [deterministic]
assert Tel  [FD= ImplementT

]
*)
*)

oops

end
