(*****************************************************************************
 * HOL-TestGen --- theorem-prover based test case generation
 *                 http://www.brucker.ch/projects/hol-testgen/
 *                                                                            
 * hoare_test_theory.thy --- testing hoare-like specifications 
 * This file is part of HOL-TestGen.
 *
 * Copyright (c) 2003-2007 ETH Zurich, Switzerland
 *               2008-2012 Achim D. Brucker, Germany
 *               2009-2012 Univ. Paris-Sud, France
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *
 *     * Neither the name of the copyright holders nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)
(* $Id: program_based_testing.thy 9448 2012-10-14 19:20:31Z brucker $ *)

header{* A Proof-Of-Technology of Program-based Testing: The Framework*}

theory 
  program_based_testing
imports
  "IMP-2011/VC"
  "IMP-2011/Denotation"
  Testing

begin  

subsection{* Unfold and its Correctness *}
text{* The core of our white box testing function
   is the following ``unwind'' function, that
   ``unfolds'' while loops and normalizes the resulting
   program in order to expose it to the operational
   semantics (i.e. the ``natural semantics'' 
   @{text evalc} up to an unwind factor k. Evaluating 
   programs leads to accumulating  path-conditions: If a 
   remaining constraint (whose components essentially result 
   from applications of the @{text If_split} rule), is satisfiable that a 
   path through a program is traceable and results to a certain
   successor state.
   
   This can be used to test program specifications:
   Hoare-Triples were checked against for all paths up to 
   a certain depth.
*}


 
primrec Append :: "[com,com] \<Rightarrow> com" (infixr "@@" 70)
where
  conc_skip : "SKIP @@ c = c"
| conc_ass  : "(x:== E) @@ c = ((x:== E); c)"
| conc_semi : "(c;d) @@ e = (c; d @@ e)"
| conc_If   : "(IF b THEN c ELSE d) @@ e = 
                       (IF b THEN c @@ e ELSE d @@ e)"
| conc_while: "(WHILE b DO c) @@ e = ((WHILE b DO c);e)"

lemma C_skip_cancel1[simp] : "C(SKIP;c) = C(c)"
  by (simp add: Denotation.C.simps Id_O_R R_O_Id)


lemma C_skip_cancel2[simp] : "C(c;SKIP) = C(c)"
  by (simp add: Denotation.C.simps Id_O_R R_O_Id)

lemma C_If_semi[simp] : 
      "C((IF x THEN c ELSE d);e) = C(IF x THEN (c;e) ELSE (d;e))"
  by auto

lemma comappend_correct [simp]: "C(c @@ d) = C(c;d)"
  apply(induct "c")
  apply(simp_all only: C_If_semi conc_If)
  apply(simp_all add: Relation.O_assoc)
  done  

fun unfold :: "nat \<times> com \<Rightarrow> com"
where
  uf_skip : "unfold(n, SKIP)     = SKIP"
| uf_ass  : "unfold(n, a :== E)  = (a :== E)"
| uf_If   : "unfold(n, IF b THEN c ELSE d) = 
             IF b THEN unfold(n, c) ELSE unfold(n, d)"
| uf_while: "unfold(n, WHILE b DO c) = 
            (if 0 < n 
             then IF b THEN unfold(n,c)@@unfold(n- 1,WHILE b DO c) ELSE SKIP
             else WHILE b DO unfold(0, c))"
| uf_semi1: "unfold(n, SKIP ; c) = unfold(n, c)"
| uf_semi2: "unfold(n, c ; SKIP) = unfold(n, c)"
| uf_semi3: "unfold(n, (IF b THEN c ELSE d) ; e) = 
            (IF b THEN (unfold(n,c;e)) ELSE (unfold(n,d;e)))"
| uf_semi4: "unfold(n, (c ; d); e) =  (unfold(n, c;d))@@(unfold(n,e))"
| uf_semi5: "unfold(n, c ; d) =  (unfold(n, c))@@(unfold(n, d))"



lemma unfold_correct_aux1 :
  assumes H : "\<forall>x. C (unfold (x, c))   = C c"
  shows       "C(unfold(n,WHILE b DO c)) = C(WHILE b DO c)"
proof (induct "n")
  case 0 then show ?case
       by(simp add: Denotation.C.simps H)
next
  case (Suc n) then show ?case
       apply(subst uf_while,subst if_P, simp)
       apply(rule_tac s = "n" and t = "Suc n - 1" in subst,arith)
       apply(simp only: Denotation.C.simps  comappend_correct)
       apply(simp only: Denotation.C.simps [symmetric] H)
       apply(simp only: Denotation.C_While_If)
       done
qed

declare uf_while [simp del]

lemma unfold_correct_aux2 : 
  "C(unfold(n,c;d))= C(unfold(n,c) ; unfold(n, d))"
proof (induct "c") print_cases
  case SKIP then show ?case by(simp)
next
  case (Assign loc E) then show ?case by(case_tac "d", simp_all)
next
  case (Semi c1 c2) then show ?case   by(case_tac "d", simp_all)
next
  case (Cond cond then_branch else_branch) then show ?case
       apply (case_tac "d", simp_all)
       apply (simp_all only: C.simps[symmetric] C_If_semi) 
       by auto
next
  case (While cond body) then show ?case by(case_tac "d", simp_all)
qed



lemma unfold_correct [rule_format]: "\<forall> x. (C(unfold(x,c)) = C(c))"
proof(induct "c")  
  case SKIP then show ?case by simp
next
  case (Assign loc E) then show ?case by simp
next
  case (Semi c1 c2) then show ?case 
        by (cases "c1", cases "c2",
            simp_all add: unfold_correct_aux2)
next
  case (Cond cond then_branch else_branch) then show ?case by simp
next
  case (While cond body) then show ?case 
        by (intro allI unfold_correct_aux1,auto)
qed


(* not really necessary, but interesting. *)
lemma wp_unfold : "wp (c) (p) = wp(unfold(n,c)) (p) "
  by(simp add: wp_def unfold_correct denotational_is_natural[symmetric])


(* not really necessary, but interesting. *)
lemma wp_test : "\<forall>\<sigma>. P \<sigma> \<longrightarrow> wp (unfold(k,c)) Q \<sigma>  \<Longrightarrow> |- {P} c {Q}"
  apply  (rule Hoare.strengthen_pre)
  apply  (simp add: wp_unfold[symmetric])
  apply  (rule wp_is_pre)
  done



subsection{* Symbolic Evaluation Rule-Set *}

lemma If_split:
 "\<lbrakk> b s \<Longrightarrow> \<langle>c0,s\<rangle> \<longrightarrow>\<^sub>c s';
   \<not> b s \<Longrightarrow> \<langle>c1,s\<rangle> \<longrightarrow>\<^sub>c s' \<rbrakk>
  \<Longrightarrow> \<langle>IF b THEN c0 ELSE c1,s\<rangle> \<longrightarrow>\<^sub>c s'"
  by (cases "b s", simp_all)

lemma If_splitE:
  "\<lbrakk>   \<langle>IF b THEN c ELSE d,s\<rangle> \<longrightarrow>\<^sub>c s';
       \<lbrakk> b s; \<langle>c,s\<rangle> \<longrightarrow>\<^sub>c s' \<rbrakk> \<Longrightarrow> P;
       \<lbrakk>  \<not> b s; \<langle>d,s\<rangle> \<longrightarrow>\<^sub>c s' \<rbrakk> \<Longrightarrow> P \<rbrakk> \<Longrightarrow> P"

  by(cases "b s", simp_all)


subsection{* Splitting Rule for program-based Tests  *}


lemma symbolic_eval_test : 
   "( |- {Pre} c {Post}) = 
      (\<forall>s t.  \<langle>unfold (n, c),s\<rangle> \<longrightarrow>\<^sub>c t \<longrightarrow> Pre s \<longrightarrow> Post t)"
proof -
    have hoare_sound_complete : "( |- {Pre} c {Post}) = ( |= {Pre} c {Post})"
         by(auto intro!: hoare_sound hoare_relative_complete)
    show ?thesis
         by(simp only: hoare_sound_complete hoare_valid_def 
                       denotational_is_natural[symmetric] unfold_correct) 
qed


subsection{* Tactic Set-up*}

ML{* 

fun contains_eval n thm =
    let fun T("Natural.evalc",_) = true | T _ = false
    in  Term.exists_Const T (term_of(cprem_of thm n)) end

*}
ML{*
local open TestGen in

fun thyp_ify_partial_evaluations ctxt = 
    (COND' contains_eval (thyp_ify ctxt) (K all_tac))

end

*}



lemmas one_point_rules = HOL.simp_thms(39) HOL.simp_thms(40)

lemma IF_split: 
"\<langle>IF b THEN c ELSE d,s\<rangle> \<longrightarrow>\<^sub>c s' = 
 ((b s \<and> \<langle>c ,s\<rangle> \<longrightarrow>\<^sub>c s')  \<or> (\<not> b s \<and> \<langle>d ,s\<rangle> \<longrightarrow>\<^sub>c s' ))"
by(cases "b s", auto)

lemma assign_sequence:
"\<langle>a:== e; c,s\<rangle> \<longrightarrow>\<^sub>c s' = \<langle>c,s[a \<mapsto> e s]\<rangle> \<longrightarrow>\<^sub>c s'"
by(simp only:Natural.semi Natural.assign one_point_rules)



lemmas symbolic_evaluation =  IF_split
                              Natural.skip Natural.assign 
                              Natural.semi  Natural.whileFalse

thm symbolic_evaluation


lemmas symbolic_evaluation2 = IF_split assign_sequence 
                              Natural.skip Natural.assign 
                              Natural.whileFalse 



lemmas memory_model = Fun.fun_upd_other HOL.simp_thms(8)
                      Fun.fun_upd_same  Fun.fun_upd_triv


ML{*

local open TestGen HOLogic in

fun generate_program_splitter ctxt simps depth no thm = 
    let val thy = theory_of_thm thm
        val  @{term "Hoare.hoare"} $ PRE $ PROG $ _ = dest_Trueprop(term_of(cprem_of thm 1));  
        val S = (Thm.trivial (cterm_of thy (mk_Trueprop
                           (@{term "Hoare.hoare"} $ PRE  $ PROG $ 
                            Free("POSTCONDITION", 
                            @{typ "(Com.loc \<Rightarrow> nat) \<Rightarrow> bool"}))))) 
        val S = S |$> (res_inst_tac ctxt (* [("n1", Int.toString depth)] *)
                                          [(("n",1),Int.toString depth)] 
                             (@{thm "symbolic_eval_test"} RS iffD2) no)
                  |$> (safe_tac ctxt)
                  |$> (asm_full_simp_tac((global_simpset_of thy) 
                             addsimps 
                            (@{thms "Append.simps"} @ 
                             @{thms unfold.simps} @
                             [@{thm uf_while}])) 1)
                  |$> (asm_full_simp_tac( HOL_ss
                             addsimps 
                            (@{thms "symbolic_evaluation2"} @ 
                             @{thms "memory_model"} @ simps @
                             [@{thm "update_def"}])) 1)
                  |$> (safe_tac ctxt)
                  |$> (ALLGOALS(COND' contains_eval (thyp_ify ctxt) (K all_tac)))

    in  thm |> (rtac (Drule.export_without_context S) 1)
  end

end (* local *)

*}

end
 
